/* --------------------------------------------------------------------------
  FILE      : nand.c                                                   
  PURPOSE   : NAND driver file
  PROJECT   : DaVinci User Boot-Loader and Flasher
  AUTHOR    : Daniel Allred
  DESC      : Generic NAND driver file for DaVinci EMIFA peripheral
-------------------------------------------------------------------------- */

// General type include
#include <tistdtypes.h>

// This module's header file 
#include "flash_nand.h"

// Nand table structure definition
#include "flash_nand_table.h"

// Device specific CSL


// Misc. utility function include
#include "flash_util.h"

// Project specific debug functionality
#include "flash_debug.h"
#include "Keystone_EMIF16_init.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/

extern const NAND_TABLE_DeviceObj NAND_TABLE_Devices[];


/************************************************************
* Local Macro Declarations                                  *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/

static volatile Uint8 *LOCAL_flashMakeAddr (Uint32 baseAddr, Uint32 offset);
static void LOCAL_flashWriteAddr (NAND_InfoHandle hNandInfo, Uint32 addr);
static void LOCAL_flashWriteCmd (NAND_InfoHandle hNandInfo, Uint32 cmd);
static void LOCAL_flashWriteBytes (NAND_InfoHandle hNandInfo, void *pSrc, Uint32 numBytes);
static void LOCAL_flashWriteAddrCycles(NAND_InfoHandle hNandInfo, Uint32 block, Uint32 page);
static void LOCAL_flashWriteAddrBytes(NAND_InfoHandle hNandInfo, Uint32 numAddrBytes, Uint32 addr);
static void LOCAL_flashWriteRowAddrBytes(NAND_InfoHandle hNandInfo, Uint32 block, Uint32 page);
static void LOCAL_flashWriteData(NAND_InfoHandle hNandInfo, Uint32 offset, Uint32 data);
static Uint32 LOCAL_flashReadData(NAND_InfoHandle hNandInfo);
static void LOCAL_flashReadBytes(NAND_InfoHandle hNandInfo, void *pDest, Uint32 numBytes);
static Uint32 LOCAL_flashGetDetails(NAND_InfoHandle hNandInfo);

// ECC functionality
static void LOCAL_flashECCStart (NAND_InfoHandle hNandInfo);
static void LOCAL_flashECCStopAndRead(NAND_InfoHandle hNandInfo, Uint32 *eccValue);
static Uint32 LOCAL_flashECCCorrection(NAND_InfoHandle hNandInfo, Uint32 *ECCold, Uint32 *ECCnew, Uint8 *data);
static void  LOCAL_flashECCManipulate(NAND_InfoHandle hNandInfo, Uint32* in, Uint32* out, NAND_ECCManipulateMethods method);

// Wait for ready signal seen at NANDFSCR
static Uint32 LOCAL_flashWaitForRdy(Uint32 timeout);

// Wait for status result from device to read good */
static Uint32 LOCAL_flashWaitForStatus(NAND_InfoHandle hNandInfo, Uint32 timeout);


/************************************************************
* Local Variable Definitions                                *
************************************************************/


/************************************************************
* Global Variable Definitions                               *
************************************************************/


/************************************************************
* Global Function Definitions                               *
************************************************************/

// Initialze NAND interface and find the details of the NAND used
NAND_InfoHandle NAND_open(Uint32 baseCSAddr)
{
  NAND_InfoHandle hNandInfo;

  DEBUG_printString("Initializing NAND flash...\r\n");
  
  // Alloc mem for temp pages
  hNandInfo = (NAND_InfoHandle) UTIL_allocMem(sizeof(NAND_InfoObj));
    
  // Set NAND flash base address
  hNandInfo->flashBase = baseCSAddr;
    
  // Get the CSOffset ( can be 0 through (DEVICE_EMIF_NUMBER_CE_REGION -1) )
  hNandInfo->CSOffset = 0;
  while (hNandInfo->CSOffset < DEVICE_EMIF_NUMBER_CE_REGION)
  {
    if ( hNandInfo->flashBase == 
         (DEVICE_EMIF_FIRST_CE_START_ADDR + 
         (DEVICE_EMIF_INTER_CE_REGION_SIZE * hNandInfo->CSOffset) ) )
    {
      break;
    }
    hNandInfo->CSOffset++;
  }

  if (hNandInfo->CSOffset == DEVICE_EMIF_NUMBER_CE_REGION)
    return NULL;
    
  // Set width
  hNandInfo->busWidth = 8;

  // Setup AEMIF registers for NAND    
  EMIF16_Regs->NANDFCTL |= (0x1 << (hNandInfo->CSOffset));        // NAND enable for CSx
  LOCAL_flashECCStart(hNandInfo);
                         
  // Send reset command to NAND
  LOCAL_flashWriteCmd( hNandInfo, NAND_RESET );
  if ( LOCAL_flashWaitForRdy(NAND_TIMEOUT) != E_PASS )
    return NULL;
  
  // Get and set device details
  if ( LOCAL_flashGetDetails(hNandInfo) != E_PASS )
    return NULL;

  return hNandInfo;
}

// Routine to read a page from NAND
Uint32 NAND_readPage(NAND_InfoHandle hNandInfo, Uint32 block, Uint32 page, Uint8 *dest)
{
  Uint32 i;
  Uint32 eccValue[4], spareValue[4];

  // Write read command
  LOCAL_flashWriteCmd(hNandInfo,NAND_LO_PAGE);
  
  // Write address bytes
  LOCAL_flashWriteAddrCycles(hNandInfo, block, page);
  
  // Additional confirm command for big_block devices
  if(hNandInfo->bigBlock)  
    LOCAL_flashWriteCmd(hNandInfo, NAND_READ_30H);

  // Wait for data to be available
  if(LOCAL_flashWaitForRdy(NAND_TIMEOUT) != E_PASS)
    return E_FAIL;

#ifdef FOUR_BIT_ECC
  for (i=0; i < hNandInfo->numOpsPerPage; i++)
  {
    LOCAL_flashECCStart(hNandInfo);
    LOCAL_flashReadBytes(hNandInfo, &dest[hNandInfo->bytesPerOp*i], hNandInfo->bytesPerOp);
    LOCAL_flashECCStopAndRead(hNandInfo, eccValue);

    // Read ECC bytes
    LOCAL_flashReadBytes(hNandInfo, &dest[hNandInfo->bytesPerPage+(hNandInfo->spareBytesPerOp*i)], hNandInfo->spareBytesPerOp);
    LOCAL_flashECCManipulate( hNandInfo,
                              (Uint32 *)&dest[hNandInfo->bytesPerPage + (hNandInfo->spareBytesPerOp*i) + 6],
                              spareValue,
                              NAND_RBL_8TO10);
    if (LOCAL_flashECCCorrection(hNandInfo,spareValue,eccValue,&dest[hNandInfo->bytesPerOp*i]) != E_PASS)
    {
      DEBUG_printString("NAND ECC failure!\r\n");
      return E_FAIL;
    }
  }

#else
  for (i=0; i < hNandInfo->numOpsPerPage; i++)
  {
    // Starting the ECC calcualtion
    LOCAL_flashECCStart(hNandInfo);

    // Actually read bytes
    LOCAL_flashReadBytes(hNandInfo, &dest[hNandInfo->bytesPerOp*i], hNandInfo->bytesPerOp);
      
    // Stop the ECC calculation and get the ECC value
    LOCAL_flashECCStopAndRead(hNandInfo, &eccValue[i]);
  }

  // Check ECCs
  for (i=0; i < hNandInfo->numOpsPerPage; i++)
  {
    LOCAL_flashReadBytes(hNandInfo, (void*)(spareValue), hNandInfo->spareBytesPerOp);
    LOCAL_flashECCManipulate(
      hNandInfo,
      (Uint32*)(spareValue + hNandInfo->ECCOffset),
      (Uint32*)(spareValue + hNandInfo->ECCOffset),
      NAND_RBL_SWAP);

    // Verify ECC values
    if(eccValue[i] != spareValue[hNandInfo->ECCOffset])
    {
      if (LOCAL_flashECCCorrection(
        hNandInfo,
        &spareValue[hNandInfo->ECCOffset],
        &eccValue[i],
        dest+(i*hNandInfo->bytesPerOp) ) != E_PASS)
      {
        DEBUG_printString("NAND ECC failure!\r\n");
        return E_FAIL;
      }
    }
  }
#endif

  // Return status check result
  return LOCAL_flashWaitForStatus(hNandInfo, NAND_TIMEOUT);
}

// Generic routine to write a page of data to NAND
Uint32 NAND_writePage(NAND_InfoHandle hNandInfo, Uint32 block, Uint32 page, Uint8 *src)
{
  Uint32 eccValue[4];
  Uint32 spareValue[4];
  Uint8 i;

  // Blank bytes
  spareValue[0] = 0xFFFFFFFF;
  spareValue[1] = 0xFFFFFFFF;
  spareValue[2] = 0xFFFFFFFF;
  spareValue[3] = 0xFFFFFFFF;

  // Make sure the NAND page pointer is at start of page
  LOCAL_flashWriteCmd(hNandInfo,NAND_LO_PAGE);

  // Write program command
  LOCAL_flashWriteCmd(hNandInfo, NAND_PGRM_START);

  // Write address bytes
  LOCAL_flashWriteAddrCycles(hNandInfo, block, page);

#ifdef FOUR_BIT_ECC
  for (i=0; i<hNandInfo->numOpsPerPage; i++)  
  {
    LOCAL_flashECCStart(hNandInfo);

    LOCAL_flashWriteBytes(hNandInfo, &src[hNandInfo->bytesPerOp*i], hNandInfo->bytesPerOp);

    LOCAL_flashECCStopAndRead(hNandInfo,eccValue);

    LOCAL_flashECCManipulate(
      hNandInfo,
      eccValue,
      (Uint32 *)(((Uint8 *)spareValue)+6),
      NAND_RBL_10TO8);

    // Writing 0xFF for 6 bytes in spare area, and the 80-bits of ECC data
    LOCAL_flashWriteBytes(hNandInfo, spareValue, hNandInfo->spareBytesPerOp);
  }
#else

  for (i=0; i<hNandInfo->numOpsPerPage; i++)
  {
    // Starting the ECC calculation
    LOCAL_flashECCStart(hNandInfo);

    // Write the bytes
    LOCAL_flashWriteBytes(hNandInfo, &src[hNandInfo->bytesPerOp*i], hNandInfo->bytesPerOp);    

    // Stopping the ECC calculation and reading the ECC value
    LOCAL_flashECCStopAndRead(hNandInfo,&eccValue[i]);
  }

  for (i=0; i<hNandInfo->numOpsPerPage; i++)
  {
    // Swap the bytes for how the ROM needs them
    LOCAL_flashECCManipulate(hNandInfo, &(eccValue[i]),&(eccValue[i]),NAND_RBL_SWAP);

    // Place the ECC values where the ROM read routine expects them    
    spareValue[hNandInfo->ECCOffset] = eccValue[i];

    // Actually write the Spare Bytes               
    LOCAL_flashWriteBytes(hNandInfo, (void*)(spareValue), hNandInfo->spareBytesPerOp);
  }
#endif          
  // Write program end command
  LOCAL_flashWriteCmd(hNandInfo, NAND_PGRM_END);

  // Wait for the device to be ready
  if (LOCAL_flashWaitForRdy(NAND_TIMEOUT) != E_PASS)
    return E_FAIL;

  // Return status check result  
  return LOCAL_flashWaitForStatus(hNandInfo, NAND_TIMEOUT);
}

// Verify data written by reading and comparing byte for byte
Uint32 NAND_verifyPage(NAND_InfoHandle hNandInfo, Uint32 block, Uint32 page, Uint8* src, Uint8* dest)
{
  Uint32 i;

  if (NAND_readPage(hNandInfo, block, page, dest) != E_PASS)
    return E_FAIL;
    
  for (i=0; i< (hNandInfo->bytesPerPage>>2); i++)
  {
    // Check for data read errors
    if ( ((Uint32 *)src)[i] != ((Uint32 *)dest)[i] )
    {
      DEBUG_printString("Data mismatch! Verification failed.\r\n");
      return E_FAIL;
    }
  }
  return E_PASS;
}

// Verify data written by reading and comparing byte for byte
Uint32 NAND_verifyBlockErased(NAND_InfoHandle hNandInfo, Uint32 block, Uint8* dest)
{
  Uint32 i,j;
  
  for (j=0; j<hNandInfo->pagesPerBlock; j++)
  {
    if (NAND_readPage(hNandInfo, block, j, dest) != E_PASS)
      return E_FAIL;
    
    for (i=0; i<(hNandInfo->bytesPerPage>>2); i++)
    {
      // Check for data read errors
      if (((Uint32 *)dest)[i] += 0xFFFFFFFF)
      {
        DEBUG_printString("Erase verification failed! Block: ");
        DEBUG_printHexInt(block);
        DEBUG_printString(" page: ");
        DEBUG_printHexInt(j);
        DEBUG_printString(". First Failing Byte: ");
        DEBUG_printHexInt(i);
        DEBUG_printString("\r\n");
        return E_FAIL;
      }
    }
  }
  return E_PASS;
}

//Global Erase NOR Flash
Uint32 NAND_globalErase(NAND_InfoHandle hNandInfo)
{
  // We don't erase block 0, and possibly some ending blocks reserved for BBT
  return NAND_eraseBlocks( hNandInfo, 1, (hNandInfo->numBlocks - 1 - NAND_NUM_BLOCKS_RESERVED_FOR_BBT) );
}

// NAND Flash erase block function
Uint32 NAND_eraseBlocks(NAND_InfoHandle hNandInfo, Uint32 startBlkNum, Uint32 blkCnt)
{  
  Uint32 i;
  Uint32 endBlkNum = startBlkNum + blkCnt - 1;
    
  // Do bounds checking
  if ( endBlkNum >= hNandInfo->numBlocks )
    return E_FAIL;
  
  // Output info about what we are doing
  DEBUG_printString("Erasing blocks ");
  DEBUG_printHexInt(startBlkNum);
  DEBUG_printString(" through ");
  DEBUG_printHexInt(endBlkNum);
  DEBUG_printString(".\r\n");

  for (i = 0; i < blkCnt; i++)
  {
    // Start erase command
    LOCAL_flashWriteCmd(hNandInfo, NAND_BERASEC1);

  // Write the row addr bytes only
    LOCAL_flashWriteRowAddrBytes(hNandInfo, (startBlkNum+i), 0);
    
    // Confirm erase command
    LOCAL_flashWriteCmd(hNandInfo, NAND_BERASEC2);

    // Wait for the device to be ready
    if (LOCAL_flashWaitForRdy(NAND_TIMEOUT) != E_PASS)
      return E_FAIL;

    // verify the op succeeded by reading status from flash
    if (LOCAL_flashWaitForStatus(hNandInfo, NAND_TIMEOUT) != E_PASS)
      return E_FAIL;
  }
  
  return E_PASS;
}

// NAND Flash unprotect command
Uint32 NAND_unProtectBlocks(NAND_InfoHandle hNandInfo, Uint32 startBlkNum, Uint32 blkCnt)
{
  Uint32 endBlkNum = startBlkNum + blkCnt - 1;

  // Do bounds checking
  if (endBlkNum >= hNandInfo->numBlocks)
    return E_FAIL;

  // Output info about what we are doing
  DEBUG_printString("Unprotecting blocks ");
  DEBUG_printHexInt(startBlkNum);
  DEBUG_printString(" through ");
  DEBUG_printHexInt(endBlkNum);
  DEBUG_printString(".\r\n");

  LOCAL_flashWriteCmd(hNandInfo, NAND_UNLOCK_START);
  LOCAL_flashWriteRowAddrBytes(hNandInfo, startBlkNum, 0);
  
  LOCAL_flashWriteCmd(hNandInfo, NAND_UNLOCK_END);
  LOCAL_flashWriteRowAddrBytes(hNandInfo, endBlkNum, 0);
    
  return E_PASS;
}

// NAND Flash protect command
void NAND_protectBlocks(NAND_InfoHandle hNandInfo)
{
  DEBUG_printString("Protecting the entire NAND flash.\n");
  LOCAL_flashWriteCmd(hNandInfo, NAND_LOCK);
}


/************************************************************
* Local Function Definitions                                *
************************************************************/

// Generic Low-level NAND access functions
static volatile Uint8 *LOCAL_flashMakeAddr (Uint32 baseAddr, Uint32 offset)
{
  return ((volatile Uint8 *) ( baseAddr + offset ));
}

static void LOCAL_flashWriteData(NAND_InfoHandle hNandInfo, Uint32 offset, Uint32 data)
{
  volatile NAND_Ptr addr;
  NAND_Data dataword;
  dataword.l = data;

  addr.cp = LOCAL_flashMakeAddr (hNandInfo->flashBase, offset);
  switch (hNandInfo->busWidth)
  {
    case FLASH_BUS_WIDTH_1_BYTE:
      *addr.cp = dataword.c;
      break;
    case FLASH_BUS_WIDTH_2_BYTES:
      *addr.wp = dataword.w;
      break;
  }
}

static void LOCAL_flashWriteCmd (NAND_InfoHandle hNandInfo, Uint32 cmd)
{
  LOCAL_flashWriteData(hNandInfo, NAND_CLE_OFFSET, cmd);
}

static void LOCAL_flashWriteAddr (NAND_InfoHandle hNandInfo, Uint32 addr)
{
  LOCAL_flashWriteData(hNandInfo, NAND_ALE_OFFSET, addr);
}

static void LOCAL_flashWriteBytes(NAND_InfoHandle hNandInfo, void* pSrc, Uint32 numBytes)
{
  volatile NAND_Ptr destAddr, srcAddr;
  Uint32 i;
  
  srcAddr.cp = (volatile Uint8*) pSrc;
  destAddr.cp = LOCAL_flashMakeAddr (hNandInfo->flashBase, NAND_DATA_OFFSET );
  switch (hNandInfo->busWidth)
  {
    case FLASH_BUS_WIDTH_1_BYTE:
      for(i=0;i<( numBytes );i++)
        *destAddr.cp = *srcAddr.cp++;
      break;
    case FLASH_BUS_WIDTH_2_BYTES:
        for(i=0;i<( numBytes >> 1);i++)
          *destAddr.wp = *srcAddr.wp++;
        break;
    }
}

static void LOCAL_flashWriteAddrCycles(NAND_InfoHandle hNandInfo, Uint32 block, Uint32 page)
{
    LOCAL_flashWriteAddrBytes(hNandInfo, hNandInfo->numColAddrBytes, 0x00000000);
    LOCAL_flashWriteRowAddrBytes(hNandInfo, block, page);
}

static void LOCAL_flashWriteAddrBytes(NAND_InfoHandle hNandInfo, Uint32 numAddrBytes, Uint32 addr)
{    
    Uint32 i;
    for (i=0; i<numAddrBytes; i++)
    {
        LOCAL_flashWriteAddr(hNandInfo, ( (addr >> (8*i) ) & 0xff) );
  }
}

static void LOCAL_flashWriteRowAddrBytes(NAND_InfoHandle hNandInfo, Uint32 block, Uint32 page)
{
    Uint32 row_addr;
  row_addr = (block << (hNandInfo->blkShift - hNandInfo->pageShift)) | page;
  LOCAL_flashWriteAddrBytes(hNandInfo, hNandInfo->numRowAddrBytes, row_addr);
}

static Uint32 LOCAL_flashReadData (NAND_InfoHandle hNandInfo)
{
  volatile NAND_Ptr addr;
  NAND_Data cmdword;
  cmdword.l = 0x0;

  addr.cp = LOCAL_flashMakeAddr (hNandInfo->flashBase, NAND_DATA_OFFSET );
  switch (hNandInfo->busWidth)
  {
      case FLASH_BUS_WIDTH_1_BYTE:
            cmdword.c = *addr.cp;
            break;
        case FLASH_BUS_WIDTH_2_BYTES:
            cmdword.w = *addr.wp;
            break;
  }
  return cmdword.l;
}

static void LOCAL_flashReadBytes(NAND_InfoHandle hNandInfo, void* pDest, Uint32 numBytes)
{
    volatile NAND_Ptr destAddr, srcAddr;
  Uint32 i;
  
  destAddr.cp = (volatile Uint8*) pDest;
  srcAddr.cp = LOCAL_flashMakeAddr (hNandInfo->flashBase, NAND_DATA_OFFSET );
  switch (hNandInfo->busWidth)
  {
    case FLASH_BUS_WIDTH_1_BYTE:
        for(i=0;i<( numBytes );i++)
          *destAddr.cp++ = *srcAddr.cp;
        break;
    case FLASH_BUS_WIDTH_2_BYTES:
        for(i=0;i<( numBytes >> 1);i++)
          *destAddr.wp++ = *srcAddr.wp;
        break;
    }
}

// Poll bit of NANDFSR to indicate ready
static Uint32 LOCAL_flashWaitForRdy(Uint32 timeout)
{
  volatile Uint32 cnt;
  cnt = timeout;

  TSC_delay_us(2);

  while( !(EMIF16_Regs->NANDFSR & NAND_NANDFSR_READY) && ((cnt--) > 0) )

  if(cnt == 0)
  {
    DEBUG_printString("NANDWaitForRdy() Timeout!\n");
    return E_FAIL;
  }

  return E_PASS;
}


// Wait for the status to be ready in NAND register
//      There were some problems reported in DM320 with Ready/Busy pin
//      not working with all NANDs. So this check has also been added.
static Uint32 LOCAL_flashWaitForStatus(NAND_InfoHandle hNandInfo, Uint32 timeout)
{
  volatile Uint32 cnt;
  Uint32 status;
  cnt = timeout;

  do
  {
    LOCAL_flashWriteCmd(hNandInfo,NAND_STATUS);
    status = LOCAL_flashReadData(hNandInfo) & (NAND_STATUS_ERROR | NAND_STATUS_BUSY);
    cnt--;
  }
  while((cnt>0) && !status);

  if(cnt == 0)
  {
    DEBUG_printString("NANDWaitForStatus() Timeout!\n");
    return E_FAIL;
  }

  return E_PASS;
}

static void LOCAL_flashECCStart (NAND_InfoHandle hNandInfo)
{
  volatile Uint32 temp;
#ifdef FOUR_BIT_ECC
  // Write appropriate bit to start ECC calcualtions (bit 12 for four bit ECC)
  EMIF16_Regs->NANDFCTL |= (1<<(12 + (hNandInfo->CSOffset)));
#else
  // Write appropriate bit to start ECC calcualtions 
  EMIF16_Regs->NANDFCTL |= (1<<(8 + (hNandInfo->CSOffset)));
#endif
  // Flush NANDFCR write (by reading another CFG register)
  temp = EMIF16_Regs->RCSR;
}

// Read the current ECC calculation and restart process
static void LOCAL_flashECCStopAndRead (NAND_InfoHandle hNandInfo, Uint32 *eccValue)
{
  volatile Uint32 temp;

  // Flush data writes (by reading CS3 data region)
  temp = *((volatile Uint32*)(((volatile Uint8*)hNandInfo->flashBase) + DEVICE_EMIF_INTER_CE_REGION_SIZE));
#ifdef FOUR_BIT_ECC
  // Read and mask appropriate (based on CSn space flash is in) ECC regsiter
  eccValue[0] = (EMIF16_Regs->NAND4BITECC1 & 0x03FF03FF);
  eccValue[1] = (EMIF16_Regs->NAND4BITECC2 & 0x03FF03FF);
  eccValue[2] = (EMIF16_Regs->NAND4BITECC3 & 0x03FF03FF);
  eccValue[3] = (EMIF16_Regs->NAND4BITECC4 & 0x03FF03FF);
#else
  // Read and mask appropriate (based on CSn space flash is in) ECC regsiter
  eccValue[0] = ((Uint32*)(&(EMIF16_Regs->NFECCCE0)))[hNandInfo->CSOffset] & hNandInfo->ECCMask;
#endif
}

// Use old (write) and new (read) ECCs to correct errors
static Uint32 LOCAL_flashECCCorrection(NAND_InfoHandle hNandInfo, Uint32 *ECCold, Uint32 *ECCnew, Uint8 *data)
{
#ifdef FOUR_BIT_ECC
  Uint8  corrState, status, numE;
  volatile Uint32 temp;
  Int32 i;
  Uint16 addOffset, corrValue;
  Uint16* syndrome10 = (Uint16 *)ECCold;

  volatile Uint32 cnt;

  // Load the syndrome10 (from 7 to 0) values 
  for(i=7;i>=0;i--)
    EMIF16_Regs->NAND4BITECCLOAD = syndrome10[i];
   
  // Read the EMIF status and version (dummy call) 
  temp = EMIF16_Regs->ERCSR;

  // Check if error is detected
  if( (EMIF16_Regs->NAND4BITECC1 | EMIF16_Regs->NAND4BITECC2 | EMIF16_Regs->NAND4BITECC3 | EMIF16_Regs->NAND4BITECC4) == 0)
  {
    return E_PASS;
  }

  // Calcuate the address if error detected
  // Set bit13 of NANDFCR
  EMIF16_Regs->NANDFCR |= 0x2000;
  
  // bit 11:8 gives the corr_state. After correction the states can
  // be 0000, 0001, 0010, 0011. All other states are work-in-progress
  // So, checking for bit 11:10
  cnt = NAND_TIMEOUT;
  do
  {
    status = (EMIF16_Regs->NANDFSR & 0xC00);
    cnt--;
  }
  while((cnt>0) && status);

  // After bit 11:10 become 00
  corrState = (EMIF16_Regs->NANDFSR & 0xF00)>>8;

  if(corrState == 1)
    return E_FAIL;
  else
  {
    // Error detected and address calculated
    // Number of errors corrected 17:16
    numE = (EMIF16_Regs->NANDFSR & 0x30000) >> 16;
    if (numE == 0)
      return E_PASS;
    else
    {
      // bit 9:0
      addOffset = 519 - (EMIF16_Regs->NANDERRADD1 & (0x3FF));
      corrValue = EMIF16_Regs->NANDERRVAL1 & (0x3FF);
      data[addOffset] ^= (Uint8)corrValue;

      if(numE == 1)
        return E_PASS;
      else
      {
        // bit 25:16
        addOffset = 519 - ( (EMIF16_Regs->NANDERRADD1 & (0x3FF0000))>>16 );
        corrValue = EMIF16_Regs->NANDERRVAL1 & (0x3FF);
        data[addOffset] ^= (Uint8)corrValue;        

        if(numE == 2)
          return E_PASS;
        else
        {
          // bit 9:0
          addOffset = 519 - (EMIF16_Regs->NANDERRADD2 & (0x3FF));
          corrValue = EMIF16_Regs->NANDERRVAL2 & (0x3FF);
          data[addOffset] ^= (Uint8)corrValue;
          
          if (numE == 3)
            return E_PASS;
          else
          {
            // bit 25:16
            addOffset = 519 - ( (EMIF16_Regs->NANDERRADD2 & (0x3FF0000)>>16) );
            corrValue = EMIF16_Regs->NANDERRVAL2 & (0x3FF);
            data[addOffset] ^= (Uint8)corrValue;
          }
        }
      }
    }    
  }
  return E_PASS;
#else
  Uint16 ECCxorVal, byteAddr, bitAddr;

  if (!hNandInfo->ECCEnable)
    return E_FAIL;

  ECCxorVal = (Uint16)  ((ECCold[0] & 0xFFFF0000) >> 16) ^  // write ECCo
                        ((ECCold[0] & 0x0000FFFF) >> 0 ) ^  // write ECCe
                        ((ECCnew[0] & 0xFFFF0000) >> 16) ^  // read ECCo 
                        ((ECCnew[0] & 0x0000FFFF) >> 0 );   // read ECCe

  if ( ECCxorVal == (0x0000FFFF & hNandInfo->ECCMask) )
  {
    // Single Bit error - can be corrected
    ECCxorVal = (Uint16) ((ECCold[0] & 0xFFFF0000) >> 16) ^ ((ECCnew[0] & 0xFFFF0000) >> 16);
    byteAddr = (ECCxorVal >> 3);
    bitAddr = (ECCxorVal & 0x7);
    data[byteAddr] ^= (0x1 << bitAddr);
    return E_PASS;
  }
  else
  {
    // Multiple Bit error - nothing we can do
    return E_FAIL;
  }
#endif
}

static void LOCAL_flashECCManipulate(NAND_InfoHandle hNandInfo, Uint32* in, Uint32* out, NAND_ECCManipulateMethods method)
{

  switch (method)
  {
#ifdef FOUR_BIT_ECC
    case NAND_RBL_8TO10:
    {
      out[0] = ((in[0]&0x000FFC00) << 6) | ((in[0]&0x000003FF)     );
      out[1] = ((in[1]&0x000000FF) <<18) | ((in[0]&0xC0000000) >>14) |
               ((in[0]&0x3FF00000) >>20);
      out[2] = ((in[1]&0x0FFC0000) >> 2) | ((in[1]&0x0003FF00) >> 8);
      out[3] = ((in[2]&0x0000FFC0) <<10) | ((in[2]&0x0000003F) << 4) |
               ((in[1]&0xF0000000) >>28);
      break;
    }
    case NAND_RBL_10TO8:
    {
      out[0] = ((in[1]&0x00030000) <<14) | ((in[1]&0x000003FF) <<20) |
               ((in[0]&0x03FF0000) >> 6) | ((in[0]&0x000003FF)     );
      out[1] = ((in[3]&0x0000000F) <<28) | ((in[2]&0x03FF0000) << 2) |
               ((in[2]&0x000003FF) << 8) | ((in[1]&0x03FC0000) >>18);
      out[2] = ((out[2]&0xFFFF0000)    ) | ((in[3]&0x03FF0000) >>10) |
               ((in[3]&0x000003F0) >> 4);
      break;
    }
#endif
    case NAND_RBL_SWAP:
    default:
    {
      Uint32 i,temp = *in;
      volatile NAND_Ptr  dataAddr, tempAddr;
    
      dataAddr.cp = LOCAL_flashMakeAddr((Uint32) out, 3);
      tempAddr.cp = LOCAL_flashMakeAddr((Uint32) &temp,0);
        
      switch (hNandInfo->busWidth)
      {
        case FLASH_BUS_WIDTH_1_BYTE:
          for(i=0; i<4; i++) *dataAddr.cp-- = *tempAddr.cp++;
          break;
        case FLASH_BUS_WIDTH_2_BYTES:
          for(i=0; i<2; i++) *dataAddr.wp-- = *tempAddr.wp++;
          break;
      }
      break;
    }
  }

}

// Get details of the NAND flash used from the id and the table of NAND devices
static Uint32 LOCAL_flashGetDetails(NAND_InfoHandle hNandInfo)
{
  Uint32 manfID,deviceID,i,j;
  
  // Issue device read ID command
  LOCAL_flashWriteCmd( hNandInfo, NAND_RDID);
  LOCAL_flashWriteAddr( hNandInfo, NAND_RDIDADD);
    
  // Read ID bytes
  manfID   = LOCAL_flashReadData( hNandInfo ) & 0xFF;
  deviceID = LOCAL_flashReadData( hNandInfo ) & 0xFF;
  j        = LOCAL_flashReadData( hNandInfo ) & 0xFF;
  j        = LOCAL_flashReadData( hNandInfo ) & 0xFF;

  i=0;
  while (NAND_TABLE_Devices[i].devID != 0x00)
  {
    if(deviceID == NAND_TABLE_Devices[i].devID)
    {
      hNandInfo->manfID = (Uint8) manfID;
      DEBUG_printString("Manufacturer ID  = ");
      DEBUG_printHexInt(hNandInfo->manfID);
      DEBUG_printString("\r\n");
    
      hNandInfo->devID = (Uint8) NAND_TABLE_Devices[i].devID;
      DEBUG_printString("Device ID        = ");
      DEBUG_printHexInt(hNandInfo->devID);
      DEBUG_printString("\r\n");
      
      hNandInfo->pagesPerBlock = NAND_TABLE_Devices[i].pagesPerBlock;
      DEBUG_printString("Pages Per Block  = ");
      DEBUG_printHexInt(hNandInfo->pagesPerBlock);
      DEBUG_printString("\r\n");
  
      hNandInfo->numBlocks = NAND_TABLE_Devices[i].numBlocks;
      DEBUG_printString("Number of Blocks = ");
      DEBUG_printHexInt(hNandInfo->numBlocks );
      DEBUG_printString("\r\n");
      
      hNandInfo->bytesPerPage = NANDFLASH_PAGESIZE(NAND_TABLE_Devices[i].bytesPerPage);
      DEBUG_printString("Bytes Per Page   = ");
      DEBUG_printHexInt(hNandInfo->bytesPerPage);
      DEBUG_printString("\r\n");
      
      break;
    }
    i++;
  }
  if (NAND_TABLE_Devices[i].devID == 0x00) return E_FAIL;

  // Assign the big_block flag
  hNandInfo->bigBlock = (hNandInfo->bytesPerPage>MAX_BYTES_PER_OP)?TRUE:FALSE;
  
  // Assign the bytes per operation value
  hNandInfo->bytesPerOp = (hNandInfo->bytesPerPage>MAX_BYTES_PER_OP)?MAX_BYTES_PER_OP:hNandInfo->bytesPerPage;
  
  // Assign the number of operations per page value
  hNandInfo->numOpsPerPage = (hNandInfo->bytesPerOp < MAX_BYTES_PER_OP)?1:(hNandInfo->bytesPerPage >> MAX_BYTES_PER_OP_SHIFT);
  
  // Assign the number of spare bytes per operation
  hNandInfo->spareBytesPerOp = hNandInfo->bytesPerOp >> SPAREBYTES_PER_OP_SHIFT;
  
  // Setup address shift values  
  j = 0;
  while( (hNandInfo->pagesPerBlock >> j) > 1)
  {
    j++;
  }
  hNandInfo->blkShift = j;
  hNandInfo->pageShift = (hNandInfo->bigBlock)?16:8;
  hNandInfo->blkShift += hNandInfo->pageShift;
  
  // Set number of column address bytes needed
  hNandInfo->numColAddrBytes = hNandInfo->pageShift >> 3;
  
  j = 0;
  while( (hNandInfo->numBlocks >> j) > 1)
  {
    j++;
  }
        
  // Set number of row address bytes needed
  if ( (hNandInfo->blkShift + j) <= 24 )
  {
      hNandInfo->numRowAddrBytes = 3 - hNandInfo->numColAddrBytes;
  }
  else if ((hNandInfo->blkShift + j) <= 32)
  {
      hNandInfo->numRowAddrBytes = 4 - hNandInfo->numColAddrBytes;
  }
  else
  {
      hNandInfo->numRowAddrBytes = 5 - hNandInfo->numColAddrBytes;
  }
  
  // Set the ECC bit mask
  hNandInfo->ECCMask = 0x00000000;
  for (j = 0; (((hNandInfo->bytesPerOp*8)>>j) > 0x1); j++)
  {
      hNandInfo->ECCMask |= (0x00010001<<j);
  }
  
  hNandInfo->ECCOffset = (hNandInfo->bigBlock)?2:0;

  hNandInfo->ECCEnable = TRUE;
                
  return E_PASS;
}

/***********************************************************
* End file                                                 *
***********************************************************/

/* --------------------------------------------------------------------------
  HISTORY
    v1.00 - DJA - 02-Nov-2007
      Initial release
    v1.10 - DJA & SG - 15-Jan-2008
      Revision to add support for 4-bit ECC used for MLC NAND on DM355 EVM
-------------------------------------------------------------------------- */

